/*
 * Copyright ? 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *
 */
/*
 *
 * Created on November 10, 2005, 11:24 PM
 */

package com.sunlabs.archsim.fleet.ships.testing;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import java.util.Hashtable;
import java.util.Vector;

/**
 *
 * @author mh193489
 */
public class PayloadSource extends Component {
    
    /** output interface terminal */
    OutputTerminal out;
    
    /** output interface terminal */
    InputTerminal ackOut;
    
    
    /**
     *
     * Data sequences for all instances of the DataSource SHIP.
     * Static lists are used in order to allow us to
     * initialize DataSources with test data before the model is created
     * using the xml scripts.
     */
    private static Hashtable<String, Vector<Payload>> sources= new Hashtable<String, Vector<Payload>>();
    

    public PayloadSource(String name, Parameters params) {
        super(name, null);
        
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
		
				director.error("Component: " + name + ", id= " + getID() 
						+ ": Attribute " + pName + " with value " + sValue
						+ " is ignored");					
		
			} // for
		} // if params != null
		        
		buildPayloadSource();
    }
    
    public PayloadSource(String name) {
        super(name, null);
        
        buildPayloadSource();
    }
    
    protected void buildPayloadSource() {
        // add a new payloadsource, if not already created indirectly by add()
        if (! sources.containsKey(name)) {
            sources.put(name, new Vector<Payload>());
        }
        
        
        // create the action
        DataSourceAction dataAction= new DataSourceAction(name+"Action", sources.get(name));
        
        // create termianals
        out= new OutputTerminal(FleetGlobal.OUT, this);
        ackOut= new InputTerminal(FleetGlobal.ACK_OUT, this);
        
        // attach terminals to the action
        dataAction.attach(out);
        dataAction.attach(ackOut);
            	
    }
    
    /*
     * Return the data container of the PayloadSource with the given name
     */
    public static Vector<Payload> get(String name) {
        if (! sources.containsKey(name)) {
            Director.getInstance().error("PayloadSource.get(): DataSource " + name + " does not exist");
            return null;
        } else {
            return sources.get(name);
            
        }
        
    }
    

    /*
     * Return a payload from a PayloadSource.
     */   
    public static Payload get(String name, int index) {
        if (! sources.containsKey(name)) {
            Director.getInstance().error("PayloadSource.get(): DataSource " + name + " does not exist");
            return null;
        } else {
            if ((index < 0) || (sources.get(name).size() <= index)) {
                Director.getInstance().error("PayloadSource()).get: index " + index + " out of bounds [0, " + sources.get(name).size() + "]");
                return null;
            } else {
                return sources.get(name).get(index);
            }
        }
    }
    
    
    /*
     *  Add a paylaod at the end of PayloadSource.
     */
    public static void add(String name, Payload payload) {
        // check if data source exists and create one if it does not
        if (! sources.containsKey(name)) {
            sources.put(name, new Vector<Payload>());
        }
        
        sources.get(name).add(payload);
        
    }
    
    /*
     * Wrap a data inside a Payload and add it at the end of PayloadSource.
     * The source and destination fields are ommited.
     * This method is used by tests which connect SHIPs directly and bypass
     * the Switch Fabric and do not require these fields.
     */
    public static void add(String name, Data data) {
        // check if data source exists and create one if it does not
        if (! sources.containsKey(name)) {
            sources.put(name, new Vector<Payload>());
        }
        
        // wrap the data inside a payload
        Payload payload= new Payload();
        payload.data= data;
        
        sources.get(name).add(payload);
        
    }
    
    
    /*
     * Delete all data from all DataSources. Functions as a reset.
     */    
    public static void clear() {
        sources.clear();        
    } 
    
        
    public Terminal getTerminal(String name) {
        if (name.equals(out.getBaseName())
        || name.equals(out.getBaseAlias()))
            return out;
        else if (name.equals(ackOut.getBaseName())
        || name.equals(ackOut.getBaseAlias()))
            return ackOut;
        else
            return null;
        
    }
    
    private class DataSourceAction extends TriggerAction {
        
        /* flag keeping track of the first activation of the action. It is
         * necessary to enable the action.
         */
        boolean initializing;
        
        /* trigger delay */
        Delay triggerDelay= TestingDefaults.DEFAULT_TRIGGER_DELAY;
        
        /* Reference to the current data srouce, for ease of use */
        Vector<Payload> dataSource;
        
        public DataSourceAction(String name, Vector<Payload> dataSource) {
            super(name);
            
            this.dataSource= dataSource;
            
            outputFunctions.put(TestingDefaults.OUT, new DataSourceFunction());
            initializing= true;
        }
        
        protected void processEnableTrigger(Object params) { }
        
        /**
         * Initially the action is disabled. In order to enable it one has to
         * call the enableStateActive() command, which has to be done after initialiation
         * (i.e. the <reset/> operation in the xml script) but before first firing request.
         * Therefore enable this action on first attemt to fire.
         */
        
        protected boolean canFire() {
            // check if first activation of this action
            if (initializing) {
                // if there is any data to output ...
                if (dataSource.size() > 0) {
                    // enable the action
                    director.schedule(this.getEnableTriggerCommand(), null, TestingDefaults.DEFAULT_TRIGGER_DELAY);
                    initializing= false;
                }
            }
            return super.canFire();
        }
        
        private class DataSourceFunction extends Function {
            
            public DataSourceFunction() {
                this("DataSourceFunction");
            }
            
            public DataSourceFunction(String name) {
                super(name);

                // set function delay
                functionDelay= TestingDefaults.DEFAULT_FUNCTION_DELAY;
            }
            
            public Object value() {
                /* see if any data available
                 * note that one element has to be left in the datasource in order to
                 * allow firing the last data item while disabling the action
                 */
                if (dataSource.size() > 1) {
                    return dataSource.remove(0);
                    
                } else {
                    // if sequence exhasted, disable this action
                    director.schedule(DataSourceAction.this.getDisableTriggerCommand(), null, DataSourceAction.this.triggerDelay);
                    // ... and generate OutOfBound data
                    return dataSource.remove(0);
                    
                }
                
            }
            
        }
    }
    
}
