/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.gasp;
/*
 * Created on Jan 13, 2005
 * 
 * @author ib27688
 *
 */

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

import com.sunlabs.archsim.base.hierarchy.SimpleEntity;
import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.globals.GlobalDefaults;
import com.sunlabs.archsim.globals.Globals;

public class Action extends SimpleEntity {
	
	protected Command notifyStateIsActiveCommand;
	protected Command notifyStateIsInactiveCommand;
	protected Command fireCommand;
	
	/** Global settings */
	protected Globals globals= Globals.getInstance();
	
	/** The set of terminals attached to the state.
	 * The set is used as we allow no duplicates.
	 */
	protected Set<Terminal> terminals;
	
	/** count the number of inputs */
	protected int numInputs;
	
	/** count the number of active inputs */
	protected int numActiveInputs;
	
	/** Contains functions that map data from states attached to
	 * input terminals to data that is passed out through output 
	 * terminals.
	 * The connection between an input terminal to an output terminal
	 * is made via the function name, which serves as a key in the
	 * outputFunctions map.
	 */ 
	
	/** The number of outputs that are scheduled for firing.
	 * Needed for subclassess of action that can be activated
	 * and de-activated. Two subsequent activations in the row
	 * must not schedule supreflupus firings.
	 */
	protected int pendingOutputs=0;
	
	protected Map<String, Function> outputFunctions;
	
	/** does this action need to journal its activity? */
	protected boolean journalAction;
	
		
	protected Action(String name) {
		super(name);
		terminals= new HashSet<Terminal>();
		outputFunctions= new HashMap<String, Function>();
		notifyStateIsActiveCommand= new StateActiveCommand();
		notifyStateIsInactiveCommand= new StateInactiveCommand();
		fireCommand= new FireCommand();
		numInputs= 0;
		numActiveInputs= 0;
		Boolean jAction= globals.booleanValue(GlobalDefaults.JOURNAL_ACTION);
		journalAction= (jAction != null)?	
							jAction.booleanValue() 
						: 	GlobalDefaults.JOURNAL_ACTION_DEF;
	} // Action constructor
	
	protected Action(String name, CompositeEntity g) {
		this(name);
		setParent(g);
	} // Action constructor
	

	/** Attach a terminal to the action.
	 * Also notify the terminal that it is attached to this state.
	 * 
	 * @param t the terminal to be attached
	 */
	public void attach(Terminal t) {
		// add terminal to the set
		terminals.add(t);	
		t.attachAction(this);
		// count the number of input terminals
		if (t instanceof InputTerminal) {
			numInputs++;
		}
	} // Attach
	
	protected class StateActiveCommand extends Command {
		public void execute(Object param) {
			// update the count of active inputs.
			// note that the state itself contains a check
			// whether it became active twice without being
			// inactive in between
			Terminal src= (Terminal)param;
			if (journalAction) {
				State state= src.getState();
				director.journal("Action " + getAlias() 
						+ ", id= " + getID() + " notifed that state "
						+ state.getAlias() + ", id= " + state.getID() 
						+ " is active");						
			}			
			if (src instanceof InputTerminal) {
				numActiveInputs++;
				// are all of the "input" states active?
				if (canFire()) {
					// it is OK to fire
					// iterate though outputs and schedule firing
					Iterator ti= terminals.iterator();
					while (ti.hasNext()) {
						Terminal t= (Terminal) (ti.next());
						if (t instanceof OutputTerminal) {
							computeAndFire((OutputTerminal)t);
						}
						else if (t instanceof InputTerminal
								&& t.getRole() == Terminal.SELF_RESETTING_INPUT) {
							// deactivate all states attached to input terminals
							t.setStateInactive();
							if (journalAction) {
								director.journal("Action " + getAlias()
										+ ", id= " + getID()
										+ ", fired on terminal " 
										+ t.getAlias() + ", id= " + t.getID() 
										+ " to set state inactive");					
							}						
						}
					}
				}
			}
//			else if (src instanceof OutputTerminal) {
//				// output fired, reduce the pending count
//				pendingOutputs--;
//			}
			
		} // execute
	} // StateActiveCommand

	protected class FireCommand extends Command {
		public void execute(Object params) {
			TerminalValuePair pair= (TerminalValuePair)params;
			Terminal t= pair.terminal;
			Object value= pair.value;

			if (journalAction) {
				director.journal("Action " + getAlias()
						+ ", id= " + getID()
						+ ", fired on terminal " 
						+ t.getAlias() + ", id= " + t.getID() 
						+ " with data " + value);					
			}
			// firing happened, notify the terminal that connects
			// this action to the state
			t.setStateActive(value);					
			
			// reduce te count of pending outputs
			pendingOutputs--;
			
		} // execute
	} // FireCommand
	
	protected void computeAndFire(OutputTerminal t) {
		// get the corresponding function
		Function fn= (Function) (outputFunctions.get(t.getBaseName()));
		// compute the output
		Object value= fn.value();
		// package the valkue and the terminal in a vector
		TerminalValuePair pair= new TerminalValuePair();
		pair.terminal=  t;
		pair.value= value;
		// schedule firing
		director.schedule(fireCommand, pair, fn.getDelay());
		// increase teh count of pending outputs
		pendingOutputs++;
	} // computeAndFire

	/**
	 * What does an action need to do when a state becomes inactive?
	 * At the moment is seems nothing needs to be done apart from 
	 * updating the count of active inputs.
	 */
	protected class StateInactiveCommand extends Command {
		public void execute(Object param) {
			Terminal src= (Terminal)param;
			if (journalAction) {
				State state= src.getState();
				director.journal("Action " + getAlias()
						+ ", id= " + getID() + " notifed that state "
						+ state.getAlias() + ", id= " + state.getID() 
						+ " is INactive");						
			}			
			if (src instanceof InputTerminal) {
				numActiveInputs--;
			}
		} // execute
	} // StateInactiveCommand
	
	/**
	 * Determine whether this action is ready to fire.
	 * TODO This can be optimized by counting the number of
	 * 		active states and comparing the count to the
	 * 		number of input states.
	 * @return
	 */
	protected boolean canFire() {
		// first check if all of the inputs are active
		// and there are no pending outputs
		boolean ready= false;
		if (numActiveInputs == numInputs && pendingOutputs == 0) { 
			// note that we assume there will always be one input terminal
			ready= true;
			// now check if all of the states really are active
			// this is a run-time error check to catch timing
			// discrepancies
			Iterator ti= terminals.iterator();
			Terminal t=null;
			while (ready && ti.hasNext()) {
				t= (Terminal) (ti.next());
				if (t instanceof InputTerminal) {
					ready = ready && t.getState().isActive();
				}
//				else if (t instanceof OutputTerminal) {
//					ready = ready && !t.getState().isActive();					
//				}
			}
			
			// hm, maybe we should check whether t= null...
			if (!ready) {
				director.fatalError("action " + getAlias() 
						+ ", id= " + getID() 
						+ " counted all its inputs as active but state "
						+ t.getState().getAlias() + ", id= "
						+ t.getState().getID() + " is not active.");
				// director.stop();
			}
		}
		else if (numActiveInputs == numInputs && pendingOutputs != 0) {
			director.fatalError("action " + getAlias() 
					+ ", id= " + getID() 
					+ " has all inputs active but still has "
					+ pendingOutputs 
					+ " pending outputs ");
		}
		return ready;
	} // can fire
	
	public boolean isAttached(Terminal t) {
		return terminals.contains(t);
	} // isAttached
	
	/** find a terminal by its base name name. return null if not foound */
	protected Terminal getTerminal(String name) {
		Terminal t= null;
		boolean found= false;
		Iterator ti= terminals.iterator();
		while (!found && ti.hasNext()) {
			Terminal next = (Terminal)ti.next();
			if (next.getBaseName().equals(name)) {
				found= true;
				t= next;
			}
		}
		
		return t;
	} // get terminal
	
	/** turn journal for this particular state on/off */
	public void setJournal(boolean journal) {
		journalAction= journal;
	} // setJournal

	
	/** Check for all attached terminals whether they are attached
	 * to this action.
	 * 
	 * @return true if all attachemnts are OK, false otherwise
	 */
	public boolean selfCheck() {
		boolean OK= true;
		// check whether terminals and the action are attached to each other
		Iterator ti= terminals.iterator();
		while (OK && ti.hasNext()) {
			Terminal t= (Terminal)(ti.next());
			OK= OK && t.isAttached(this);
			if (!OK) {
				director.error("Action " + getAlias() + ", id= " + getID() 
						+ ", thinks it is attached to terminal " + t.getAlias()
						+ ", id= " + t.getID() + ", but the terminal is"
						+ " not attached to the action");
			}
		}
		
		// now check if every output terminal has an
		// associated function
		ti= terminals.iterator();
		while (OK && ti.hasNext()) {
			Terminal t= (Terminal)(ti.next());
			if (t instanceof OutputTerminal) {
				OK= OK && outputFunctions.containsKey(t.getBaseName());
				if (!OK) {
					director.error("Action " + getAlias() + ", id= " + getID()
							+ ", has output terminal " + 
							t.getAlias() + ", id= " + t.getID()
							+ ", but the base name " + t.getBaseAlias()
							+ " is not attached to a function.");
				}
			}
		}
		
		/** TODO temporary
		Iterator fi= outputFunctions.keySet().iterator();
		director.info("Action " + getName() + ", id= " + getID()
				+ " contains the following name function pairs:");
		while ( fi.hasNext()) {
			String name= (String) fi.next();
			Function f= (Function) outputFunctions.get(name);
			director.info("key= " + name + " function= " + f.getName());
		}
		*/
		
		return OK;
	} // selfCheck
	
	
	public void print() {
		director.info("Action " + getAlias()
				+ " AKA " + getName() 
				+ ", id= " + getID() 
				+ ", is attached to terminals: ");
		Iterator ti= terminals.iterator();
		while (ti.hasNext()) {
			Terminal t= (Terminal)(ti.next());
			director.info("\t"+ t.getAlias()
					+ " AKA " + t.getName()
					+ ", id= " + t.getID());
		}
		director.info("\t"+ "and has functions");
		Set functions= outputFunctions.entrySet();
		Iterator fi= functions.iterator();
		while (fi.hasNext()) {
			Map.Entry entry= (Map.Entry)(fi.next());
			Function f= (Function)(entry.getValue());
			String n= (String)(entry.getKey());
			director.info("\t" + "function " + f.getName()
					+ " connected to terminal " + n);
		}
		
	} // printConnections
	
	/** entity initialization - nothing needed for a generic action */
	public void init() {}


} // class Action

